#include "pnm-manipulation.h"
#include "pnm-type.h"
#include "utest-manipulation.h"

#include <CUnit/CUnit.h>

#define TPW 4
#define TPH 6
#define TPM ((TPW-1)*(TPH-1))
#define RED(X,Y)   ((TPH-1)*(X))
#define GRN(X,Y)   ((TPW-1)*(Y))
#define BLU(X,Y)   ((TPM-TPH-TPW-2)+(X)+(Y))
#define RGB(R,G,B) (rgbpic_pixel_t){(R),(G),(B)}

static rgbpic_t *test_ref = NULL, *test_work = NULL;

static inline int pixels_equal(rgbpic_pixel_t lhs, rgbpic_pixel_t rhs) {
    return lhs.red == rhs.red && lhs.green == rhs.green && lhs.blue == rhs.blue;
}

static int pics_equal(rgbpic_t *lhs, rgbpic_t *rhs) {
    if (lhs == NULL || rhs == NULL) {
        return lhs == rhs;
    }
    if (lhs->width != rhs->width || lhs->height != rhs->height || lhs->maxpixvalue != rhs->maxpixvalue) {
        return 0;
    }
    if (lhs->pixelvalues == rhs->pixelvalues) {
        return 1;
    }
    for (int y = 0; y < lhs->height; ++y) {
        for (int x = 0; x < lhs->width; ++x) {
            if (!pixels_equal(lhs->pixelvalues[rgbpic_coord(lhs, x, y)], rhs->pixelvalues[rgbpic_coord(rhs, x, y)])) {
                return 0;
            }
        }
    }
    return 1;
}

static void test_rgbpic_invert() {
    rgbpic_invert(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(TPM-RED(x,y),TPM-GRN(x,y),TPM-BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_invert_R() {
    rgbpic_invert_R(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(TPM-RED(x,y),GRN(x,y),BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_invert_G() {
    rgbpic_invert_G(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(RED(x,y),TPM-GRN(x,y),BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_invert_B() {
    rgbpic_invert_B(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(RED(x,y),GRN(x,y),TPM-BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_mirror_h() {
    rgbpic_mirror_h(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, TPW - x - 1, y)] = RGB(RED(x,y),GRN(x,y),BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_mirror_v() {
    rgbpic_mirror_v(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, TPH - y - 1)] = RGB(RED(x,y),GRN(x,y),BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_rotate_l() {
    rgbpic_rotate_l(test_work);
    test_ref = rgbpic_create(TPH, TPW, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, y, TPW - 1 - x)] = RGB(RED(x,y),GRN(x,y),BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_mirror_c() {
    rgbpic_mirror_c(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, TPW - x - 1, TPH - y - 1)] = RGB(RED(x,y),GRN(x,y),BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_rotate_r() {
    rgbpic_rotate_r(test_work);
    test_ref = rgbpic_create(TPH, TPW, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, TPH - 1 - y, x)] = RGB(RED(x,y),GRN(x,y),BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_color_RG() {
    rgbpic_color_RG(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(GRN(x,y),RED(x,y),BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_color_GB() {
    rgbpic_color_GB(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(RED(x,y),BLU(x,y),GRN(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_color_BR() {
    rgbpic_color_BR(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(BLU(x,y),GRN(x,y),RED(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_delete_R() {
    rgbpic_delete_R(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(0,GRN(x,y),BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_delete_G() {
    rgbpic_delete_G(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(RED(x,y),0,BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_delete_B() {
    rgbpic_delete_B(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(RED(x,y),GRN(x,y),0);
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_convert_to_grey_avg() {
    rgbpic_convert_to_grey_avg(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            rgbpic_monochrome_pixel_t v = (RED(x,y) + GRN(x,y) + BLU(x,y)) / 3;
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(v,v,v);
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_convert_to_grey_lum() {
    rgbpic_convert_to_grey_lum(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            rgbpic_monochrome_pixel_t v = (30 * RED(x,y) + 59 * GRN(x,y) + 11 * BLU(x,y)) / 100;
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(v,v,v);
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_convert_R_to_grey() {
    rgbpic_convert_R_to_grey(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(RED(x,y),RED(x,y),RED(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_convert_G_to_grey() {
    rgbpic_convert_G_to_grey(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(GRN(x,y),GRN(x,y),GRN(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_convert_B_to_grey() {
    rgbpic_convert_B_to_grey(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(BLU(x,y),BLU(x,y),BLU(x,y));
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_convert_to_bw_avg() {
    rgbpic_convert_to_bw_avg(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            rgbpic_monochrome_pixel_t v = ((RED(x,y) + GRN(x,y) + BLU(x,y)) * 2 < TPM * 3) ? 0 : TPM;
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(v,v,v);
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

static void test_rgbpic_convert_to_bw_lum() {
    rgbpic_convert_to_bw_lum(test_work);
    test_ref = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            rgbpic_monochrome_pixel_t v = ((30 * RED(x,y) + 59 * GRN(x,y) + 11 * BLU(x,y)) / 50 <= TPM) ? 0 : TPM;
            test_ref->pixelvalues[rgbpic_coord(test_ref, x, y)] = RGB(v,v,v);
        }
    }
    CU_ASSERT_TRUE(pics_equal(test_work, test_ref));
    rgbpic_delete(test_ref);
}

CU_TestInfo manipulation_tests[] = {
    {"invert",                                 test_rgbpic_invert},
    {"invert only red",                        test_rgbpic_invert_R},
    {"invert only green",                      test_rgbpic_invert_G},
    {"invert only blue",                       test_rgbpic_invert_B},
    {"mirror horizontally",                    test_rgbpic_mirror_h},
    {"mirror vertically",                      test_rgbpic_mirror_v},
    {"rotate 90",                              test_rgbpic_rotate_l},
    {"rotate 180",                             test_rgbpic_mirror_c},
    {"rotate 270",                             test_rgbpic_rotate_r},
    {"exchange colors red and green",          test_rgbpic_color_RG},
    {"exchange colors green and blue",         test_rgbpic_color_GB},
    {"exchange colors blue and red",           test_rgbpic_color_BR},
    {"delete red color values",                test_rgbpic_delete_R},
    {"delete green color values",              test_rgbpic_delete_G},
    {"delete blue color values",               test_rgbpic_delete_B},
    {"convert to gray numerically",            test_rgbpic_convert_to_grey_avg},
    {"convert to gray as visible",             test_rgbpic_convert_to_grey_lum},
    {"convert only red channel to gray",       test_rgbpic_convert_R_to_grey},
    {"convert only green channel to gray",     test_rgbpic_convert_G_to_grey},
    {"convert only blue channel to gray",      test_rgbpic_convert_B_to_grey},
    {"convert to black and white numerically", test_rgbpic_convert_to_bw_avg},
    {"convert to black and white as visible",  test_rgbpic_convert_to_bw_lum},
    CU_TEST_INFO_NULL
};

void manip_test_setup() {
    test_work = rgbpic_create(TPW, TPH, TPM);
    for (int y = 0; y < TPH; ++y) {
        for (int x = 0; x < TPW; ++x) {
            test_work->pixelvalues[rgbpic_coord(test_work, x, y)] = RGB(RED(x,y),GRN(x,y),BLU(x,y));
        }
    }
}

void manip_test_teardown() {
    rgbpic_delete(test_work);
}
